package LinkedList;

import static org.junit.Assert.*;

import org.junit.Test;
import java.util.Iterator;

public class CLLTest {

	@Test
    public void testConstructor() {
        CLL<Integer> c= new CLL<>();
        assertEquals("[]", c.toString());
        assertEquals("[]", c.toStringRev());
        assertEquals(0, c.size());
    }

    @Test
    /** Test append(E val) (and toStringRev() and size()) */
    public void testAppend() {
        CLL<Integer> c= new CLL<>();

        c.append(5);
        assertEquals("[5]", c.toString());
        assertEquals("[5]", c.toStringRev());
        assertEquals(1, c.size());

        c.append(17);
        assertEquals("[5, 17]", c.toString());
        assertEquals("[17, 5]", c.toStringRev());
        assertEquals(2, c.size());

        c.append(6);
        assertEquals("[5, 17, 6]", c.toString());
        assertEquals("[6, 17, 5]", c.toStringRev());
    }

    @Test
    /** Test prepend(E val) (and toStringRev() and size()) */
    public void testPrepend() {
        CLL<Integer> c= new CLL<>();

        c.prepend(5);
        assertEquals("[5]", c.toString());
        assertEquals("[5]", c.toStringRev());
        assertEquals(1, c.size());

        c.prepend(17);
        assertEquals("[17, 5]", c.toString());
        assertEquals("[5, 17]", c.toStringRev());
        assertEquals(2, c.size());

        c.prepend(6);
        assertEquals("[6, 17, 5]", c.toString());
        assertEquals("[5, 17, 6]", c.toStringRev());
    }

    @Test
    /** Test getNode(int h). */
    public void testGetNode() {
        CLL<Integer> c= new CLL<>();
        // Test throw an exception when arg is less than 0
        try {c.getNode(-1); fail();}
        catch (IndexOutOfBoundsException e) {}
        catch (Exception e) {fail();}

        c.append(5);
        c.append(17);
        c.append(22);
        c.append(24);
        c.append(2);

        assertEquals(5, (int) c.getNode(0).getValue());
        assertEquals(17, (int) c.getNode(1).getValue());
        assertEquals(22, (int) c.getNode(2).getValue());
        assertEquals(24, (int)c.getNode(3).getValue());
        assertEquals(2, (int)c.getNode(4).getValue());

        // Test throw an exception when arg is less than 0
        try {c.getNode(-1); fail();}
        catch (IndexOutOfBoundsException e) {}
        catch (Exception e) {fail();}

        // Test throw an exception when arg is at least size of the list
        try {c.getNode(5); fail();}
        catch (IndexOutOfBoundsException e) {}
        catch (Exception e) {fail();}
    }

    @Test
    /** Test insertAfter(E v, Node n) */
    public void testInsertAfter() {
        CLL<Integer> c= new CLL<>();
		c.append(5);
        c.insertAfter(5, c.getNode(0));
        assertEquals("[5, 5]", c.toString());
        assertEquals("[5, 5]", c.toStringRev());
        assertEquals(2, c.size());

        c.insertAfter(22, c.getNode(0));
        assertEquals("[5, 22, 5]", c.toString());
        assertEquals("[5, 22, 5]", c.toStringRev());
        assertEquals(3, c.size());

        c.insertAfter(17, c.getNode(1));
        assertEquals("[5, 22, 17, 5]", c.toString());
        assertEquals("[5, 17, 22, 5]", c.toStringRev());
        assertEquals(4, c.size());

        c.insertAfter(15, c.getNode(1));
        assertEquals("[5, 22, 15, 17, 5]", c.toString());
        assertEquals("[5, 17, 15, 22, 5]", c.toStringRev());
        assertEquals(5, c.size());

        c.insertAfter(6, c.getNode(4));
        assertEquals("[5, 22, 15, 17, 5, 6]", c.toString());
        assertEquals("[6, 5, 17, 15, 22, 5]", c.toStringRev());
        assertEquals(6, c.size());
    }

    @Test
    public void testRemove() {
        CLL<Integer> c= new CLL<>();

        c.append(5);
        // Delete only node
        c.remove(c.getNode(0));
        assertEquals("[]", c.toString());
        assertEquals("[]", c.toStringRev());
        assertEquals(0, c.size());

        c.append(17);
        c.append(22);
        // Delete first node when more than one
        c.remove(c.getNode(0));
        assertEquals("[22]", c.toString());
        assertEquals("[22]", c.toStringRev());
        assertEquals(1, c.size());

        c.append(25);
        // Delete last node when more than one
        c.remove(c.getNode(1));
        assertEquals("[22]", c.toString());
        assertEquals("[22]", c.toStringRev());
        assertEquals(1, c.size());

        c.append(24);
        c.append(27);
        assertEquals("[22, 24, 27]", c.toString());
        // Delete middle node
        c.remove(c.getNode(1));
        assertEquals("[22, 27]", c.toString());
        assertEquals("[27, 22]", c.toStringRev());
        assertEquals(2, c.size());
    }

    @Test
    public void testClassOtherThanInteger() {
        CLL<String> c= new CLL<>();
        c.append("L");
        c.append("A");
        c.append(null);
        c.append("T");
        c.append("T");
        c.append("E");
        assertEquals("[L, A, null, T, T, E]", c.toString());
    }


    /*************************************************************************/

    @Test
    public void testIterator() {
        CLL<Integer> cll= new CLL<>();
        assertEquals("", useIterator(cll));

        cll.append(1); cll.append(8); cll.append(7);
        cll.append(5); cll.append(4); cll.append(3);
        assertEquals("1 8 7 5 4 3 ", useIterator(cll));
    }

    /** Return the enumeration of cll with a blank after each value,
     * using cll.iterator() */
    public static <V> String useIterator(CLL<V> cll) {
        String res= "";
        Iterator<V> it= cll.iterator();
        // invariant: res contains the values that have been
        //        enumerated so far, with a blank after each
        while (it.hasNext()) {
            res= res + it.next() + " ";
        }
        return res;
    }

    @Test
    public void testIterable() {
        CLL<Integer> cll= new CLL<>();
        assertEquals("", useForEach(cll));

        cll.append(1); cll.append(8); cll.append(7);
        cll.append(5); cll.append(4); cll.append(3);
        assertEquals("1 8 7 5 4 3 ", useForEach(cll));
    }

    /** Return the enumeration of cll, with a blank after each value */
    public static <V> String useForEach(CLL<V> cll) {
        String res= "";
        // invariant: res contains the values that have been
        //        enumerated so far (with a blank after each)
        for (V elem : cll) {
            res= res + elem + " ";
        }
        return res;
    }
}